@@ -23,7 +23,6 @@
  */
 #include "hw.h"
 #include "pc.h"
-#include "apic.h"
 #include "isa.h"
 #include "monitor.h"
 #include "qemu-timer.h"
@@ -192,6 +191,7 @@ int64_t irq_time[16];
 static void i8259_set_irq(void *opaque, int irq, int level)
 {
     PicState2 *s = opaque;
+
 #if defined(DEBUG_PIC) || defined(DEBUG_IRQ_COUNT)
     if (level != irq_level[irq]) {
         DPRINTF("i8259_set_irq: irq=%d level=%d\n", irq, level);
@@ -204,7 +204,7 @@ static void i8259_set_irq(void *opaque, int irq, int level)
 #endif
 #ifdef DEBUG_IRQ_LATENCY
     if (level) {
-        irq_time[irq] = qemu_get_clock(vm_clock);
+        irq_time[irq] = qemu_get_clock_ns(vm_clock);
     }
 #endif
     pic_set_irq1(&s->pics[irq >> 3], irq & 7, level);
@@ -220,14 +220,13 @@ static inline void pic_intack(PicState *s, int irq)
     } else {
         s->isr |= (1 << irq);
     }
-
     /* We don't clear a level sensitive interrupt here */
     if (!(s->elcr & (1 << irq)))
         s->irr &= ~(1 << irq);
-
 }
 
 extern int time_drift_fix;
+extern int64_t timer_acks, timer_ints_to_push;
 
 int pic_read_irq(PicState2 *s)
 {
@@ -235,11 +234,9 @@ int pic_read_irq(PicState2 *s)
 
     irq = pic_get_irq(&s->pics[0]);
     if (irq >= 0) {
-
         pic_intack(&s->pics[0], irq);
 #ifdef TARGET_I386
 	if (time_drift_fix && irq == 0) {
-	    extern int64_t timer_acks, timer_ints_to_push;
 	    timer_acks++;
 	    if (timer_ints_to_push > 0) {
 		timer_ints_to_push--;
@@ -274,13 +271,15 @@ int pic_read_irq(PicState2 *s)
 #ifdef DEBUG_IRQ_LATENCY
     printf("IRQ%d latency=%0.3fus\n",
            irq,
-           (double)(qemu_get_clock(vm_clock) -
+           (double)(qemu_get_clock_ns(vm_clock) -
                     irq_time[irq]) * 1000000.0 / get_ticks_per_sec());
 #endif
     DPRINTF("pic_interrupt: irq=%d\n", irq);
     return intno;
 }
 
+static int kvm_kernel_pic_load_from_user(PicState *s);
+
 static void pic_reset(void *opaque)
 {
     PicState *s = opaque;
@@ -301,6 +300,10 @@ static void pic_reset(void *opaque)
     s->init4 = 0;
     s->single_mode = 0;
     /* Note: ELCR is not reset */
+
+    if (kvm_enabled() && kvm_irqchip_in_kernel()) {
+        kvm_kernel_pic_load_from_user(s);
+    }
 }
 
 static void pic_ioport_write(void *opaque, uint32_t addr, uint32_t val)
@@ -468,7 +471,6 @@ static uint32_t elcr_ioport_read(void *opaque, uint32_t addr1)
 }
 
 static void kvm_kernel_pic_save_to_user(PicState *s);
-static int kvm_kernel_pic_load_from_user(PicState *s);
 
 static void pic_pre_save(void *opaque)
 {
@@ -590,7 +592,7 @@ static void kvm_kernel_pic_save_to_user(PicState *s)
     chip.chip_id = (&s->pics_state->pics[0] == s) ?
                    KVM_IRQCHIP_PIC_MASTER :
                    KVM_IRQCHIP_PIC_SLAVE;
-    kvm_get_irqchip(kvm_context, &chip);
+    kvm_get_irqchip(kvm_state, &chip);
     kpic = &chip.chip.pic;
 
     s->last_irr = kpic->last_irr;
@@ -640,7 +642,7 @@ static int kvm_kernel_pic_load_from_user(PicState *s)
     kpic->elcr = s->elcr;
     kpic->elcr_mask = s->elcr_mask;
 
-    kvm_set_irqchip(kvm_context, &chip);
+    kvm_set_irqchip(kvm_state, &chip);
 #endif
     return 0;
 }
