@@ -25,36 +25,9 @@
 register struct CPUSH4State *env asm(AREG0);
 
 #include "cpu.h"
-#include "exec-all.h"
-
-static inline int cpu_has_work(CPUState *env)
-{
-    return (env->interrupt_request & CPU_INTERRUPT_HARD);
-}
-
-static inline int cpu_halted(CPUState *env) {
-    if (!env->halted)
-        return 0;
-    if (cpu_has_work(env)) {
-        env->halted = 0;
-        env->intr_at_halt = 1;
-        return 0;
-    }
-    return EXCP_HALTED;
-}
 
 #ifndef CONFIG_USER_ONLY
 #include "softmmu_exec.h"
 #endif
 
-static inline void regs_to_env(void)
-{
-    /* XXXXX */
-}
-
-static inline void env_to_regs(void)
-{
-    /* XXXXX */
-}
-
 #endif				/* _EXEC_SH4_H */
